import logging
import os
import threading
import time

from urllib3.util.retry import Retry

try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin

import requests

_LOG = logging.getLogger("ubipop")

HTTP_TOTAL_RETRIES = int(os.environ.get("UBIPOP_HTTP_TOTAL_RETRIES", 10))
HTTP_RETRY_BACKOFF = float(os.environ.get("UBIPOP_HTTP_RETRY_BACKOFF", 1))
HTTP_TIMEOUT = int(os.environ.get("UBIPOP_HTTP_TIMEOUT", 120))


class UnsupportedTypeId(Exception):
    pass


class PulpRetryAdapter(requests.adapters.HTTPAdapter):
    def __init__(self, *args, **kwargs):
        kwargs["max_retries"] = Retry(
            total=kwargs.get("total_retries", HTTP_TOTAL_RETRIES),
            status_forcelist=[500, 502, 503, 504],
            method_whitelist=[
                "HEAD",
                "TRACE",
                "GET",
                "POST",
                "PUT",
                "OPTIONS",
                "DELETE",
            ],
            backoff_factor=kwargs.get("backoff_factor", HTTP_RETRY_BACKOFF),
        )
        super(PulpRetryAdapter, self).__init__(*args, **kwargs)


class Pulp(object):
    PULP_API = "/pulp/api/v2/"

    def __init__(self, hostname, **kwargs):
        self.hostname = hostname
        self.scheme = "https://" if not hostname.startswith("https://") else ""
        self.base_url = urljoin(self.scheme + hostname, self.PULP_API)
        self.local = threading.local()

        self._session_kwargs = {}

        if kwargs.get("verify") is False:
            import urllib3

            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        for arg in (
            "auth",
            "cert",
            "verify",
        ):
            if arg in kwargs:
                self._session_kwargs[arg] = kwargs.pop(arg)

    def _make_session(self):
        adapter = PulpRetryAdapter()
        session = requests.Session()
        session.mount("http://", adapter)
        session.mount("https://", adapter)

        for key, value in self._session_kwargs.items():
            setattr(session, key, value)

        self.local.session = session

    def do_request(self, req_type, url, data=None):
        if not hasattr(self.local, "session"):
            self._make_session()

        req_url = urljoin(self.base_url, url)

        if req_type == "post":
            ret = self.local.session.post(req_url, json=data, timeout=HTTP_TIMEOUT)
        elif req_type == "get":
            ret = self.local.session.get(req_url, timeout=HTTP_TIMEOUT)
        else:
            ret = None

        return ret

    def wait_for_tasks(self, task_id_list, delay=5.0):
        results = {}

        _tasks = set(task_id_list)
        while _tasks:
            statuses = self.search_tasks(_tasks)
            for status in statuses:
                if status["state"] in ("finished", "error", "canceled"):
                    _tasks -= set([status["task_id"]])
                results[status["task_id"]] = status
            if _tasks:
                time.sleep(delay)
        return results

    def search_tasks(self, task_ids):
        url = "tasks/{task_id}/"
        statuses = []
        for task_id in task_ids:
            ret = self.do_request("get", url.format(task_id=task_id))
            statuses.append(ret.json())
        return statuses

    def _modules_query(self, modules):
        query_list = []
        for module in modules:
            query_list.append(
                {
                    "$and": [
                        {"name": module.name},
                        {"context": module.context},
                        {"version": module.version},
                        {"stream": module.stream},
                        {"arch": module.arch},
                    ]
                }
            )

        return query_list

    def _module_defaults_query(self, module_defaults):
        query_list = []
        for md_d in module_defaults:
            query_list.append({"$and": [{"name": md_d.name}, {"stream": md_d.stream}]})
        return query_list

    def _rpms_query(self, rpms):
        return [{"filename": rpm.filename} for rpm in rpms]

    def _filter_item_log(self, type_ids, unit):
        log_message = ""
        if "rpm" in type_ids or "srpm" in type_ids:
            log_message = unit.filename
        elif "modulemd" in type_ids:
            log_message = unit.nsvca
        elif "modulemd_defaults" in type_ids:
            log_message = ":".join((unit.name, unit.stream))
        return log_message

    def unassociate_units(self, repo, units, type_ids):
        url = "repositories/{dst_repo}/actions/unassociate/".format(dst_repo=repo.id)
        data = {
            "criteria": {
                "type_ids": list(type_ids),
                "filters": {"unit": {"$or": self._get_query_list(type_ids, units)}},
            },
        }
        log_msg = "Unassociating %s(%s) from %s"
        for unit in units:
            info = self._filter_item_log(type_ids, unit)
            _LOG.info(log_msg, ",".join(type_ids), info, repo.id)
        ret = self.do_request("post", url, data).json()
        return [task["task_id"] for task in ret["spawned_tasks"]]

    def associate_units(self, src_repo, dest_repo, units, type_ids):
        url = "repositories/{dst_repo}/actions/associate/".format(dst_repo=dest_repo.id)
        data = {
            "source_repo_id": src_repo.id,
            "criteria": {
                "type_ids": list(type_ids),
                "filters": {
                    "unit": {
                        "$or": self._get_query_list(type_ids, units),
                    },
                },
            },
        }
        log_msg = "Associating %s(%s) from %s to %s"
        for unit in units:
            info = self._filter_item_log(type_ids, unit)
            _LOG.info(log_msg, ",".join(type_ids), info, src_repo.id, dest_repo.id)
        ret = self.do_request("post", url, data)
        ret.raise_for_status()
        ret_json = ret.json()
        return [task["task_id"] for task in ret_json["spawned_tasks"]]

    def _get_query_list(self, type_ids, units):
        if "modulemd" in type_ids:
            query_list = self._modules_query(units)
        elif "modulemd_defaults" in type_ids:
            query_list = self._module_defaults_query(units)
        elif "rpm" in type_ids or "srpm" in type_ids:
            query_list = self._rpms_query(units)
        else:
            raise UnsupportedTypeId

        return query_list

    def associate_modules(self, src_repo, dst_repo, modules):
        return self.associate_units(src_repo, dst_repo, modules, ("modulemd",))

    def associate_module_defaults(self, src_repo, dst_repo, module_defaults):
        return self.associate_units(
            src_repo, dst_repo, module_defaults, ("modulemd_defaults",)
        )

    def associate_packages(self, src_repo, dst_repo, rpms):
        return self.associate_units(src_repo, dst_repo, rpms, ("rpm", "srpm"))

    def unassociate_modules(self, repo, modules):
        return self.unassociate_units(repo, modules, ("modulemd",))

    def unassociate_module_defaults(self, repo, module_defaults):
        return self.unassociate_units(repo, module_defaults, ("modulemd_defaults",))

    def unassociate_packages(self, repo, rpms):
        return self.unassociate_units(repo, rpms, ("rpm", "srpm"))



import attr


@attr.s
class UbiManifest(object):
    repo_id = attr.ib()
    manifest = attr.ib()

    @classmethod
    def from_data(cls, data):
        repo_id = data["repo_id"]
        items = []
        for item in data["content"]:
            new_item = cls.make_ubi_unit(item, repo_id)
            if new_item:
                items.append(new_item)

        return cls(repo_id=repo_id, manifest=items)

    @staticmethod
    def make_ubi_unit(manifest_item, dst_repo_id):
        out = None

        unit_type = manifest_item["unit_type"]
        src_repo_id = manifest_item["src_repo_id"]
        value = manifest_item["value"]

        if unit_type == "RpmUnit":
            out = RpmUnit(src_repo_id, dst_repo_id, value)
        elif unit_type == "ModulemdUnit":
            out = ModulemdUnit.from_nsvca(value, src_repo_id, dst_repo_id)
        elif unit_type == "ModulemdDefaultsUnit":
            out = ModulemdDefaultsUnit.from_ns(value, src_repo_id, dst_repo_id)
        else:
            # unknown unit
            pass

        return out

    @property
    def packages(self):
        return self._filter_units_by_class(RpmUnit)

    @property
    def modules(self):
        return self._filter_units_by_class(ModulemdUnit)

    @property
    def modulemd_defaults(self):
        return self._filter_units_by_class(ModulemdDefaultsUnit)

    def _filter_units_by_class(self, klass):
        return [unit for unit in self.manifest if isinstance(unit, klass)]


@attr.s
class UbiCompatibleUnit(object):
    src_repo_id = attr.ib()
    dst_repo_id = attr.ib()

    @property
    def associate_source_repo_id(self):
        return self.src_repo_id


@attr.s
class RpmUnit(UbiCompatibleUnit):
    filename = attr.ib()


@attr.s
class ModulemdUnit(UbiCompatibleUnit):
    name = attr.ib()
    stream = attr.ib()
    version = attr.ib()
    context = attr.ib()
    arch = attr.ib()

    @classmethod
    def from_nsvca(cls, nsvca, src_repo_id, dst_repo_id):
        n, s, v, c, a = nsvca.split(":")
        return cls(
            name=n,
            stream=s,
            version=int(v),
            context=c,
            arch=a,
            src_repo_id=src_repo_id,
            dst_repo_id=dst_repo_id,
        )

    @property
    def nsvca(self):
        return ":".join(
            (self.name, self.stream, str(self.version), self.context, self.arch)
        )


@attr.s
class ModulemdDefaultsUnit(UbiCompatibleUnit):
    name = attr.ib()
    stream = attr.ib()

    @classmethod
    def from_ns(cls, ns, src_repo_id, dst_repo_id):
        n, s = ns.split(":")

        return cls(name=n, stream=s, src_repo_id=src_repo_id, dst_repo_id=dst_repo_id)



class PulpAction(object):
    def __init__(self, units, dst_repo):
        self.units = units
        self.dst_repo = dst_repo

    def get_actions(self, pulp_client_inst):
        raise NotImplementedError


class AssociateAction(PulpAction):
    def __init__(self, units, dst_repo, src_repos):
        super(AssociateAction, self).__init__(units, dst_repo)
        self.src_repos = src_repos

    def _map_src_repo_to_unit(self):
        src_repo_unit_map = {}
        for unit in self.units:
            src_repo_unit_map.setdefault(unit.associate_source_repo_id, []).append(unit)

        return src_repo_unit_map

    def _get_repo_obj(self, repo_id):
        for repo in self.src_repos:
            if repo_id == repo.id:
                return repo

    def get_actions(self, pulp_client_inst):
        raise NotImplementedError


class AssociateActionModules(AssociateAction):
    TYPE = "modules"

    def get_actions(self, pulp_client_inst):
        actions = []
        for src_repo_id, units in self._map_src_repo_to_unit().items():
            actions.append(
                (
                    pulp_client_inst.associate_modules,
                    self._get_repo_obj(src_repo_id),
                    self.dst_repo,
                    units,
                )
            )

        return actions


class UnassociateActionModules(PulpAction):
    TYPE = "modules"

    def get_actions(self, pulp_client_inst):
        return [(pulp_client_inst.unassociate_modules, self.dst_repo, self.units)]


class AssociateActionModuleDefaults(AssociateAction):
    TYPE = "module_defaults"

    def get_actions(self, pulp_client_inst):
        actions = []
        for src_repo_id, units in self._map_src_repo_to_unit().items():
            actions.append(
                (
                    pulp_client_inst.associate_module_defaults,
                    self._get_repo_obj(src_repo_id),
                    self.dst_repo,
                    units,
                )
            )

        return actions


class UnassociateActionModuleDefaults(PulpAction):
    TYPE = "module_defaults"

    def get_actions(self, pulp_client_inst):
        return [
            (pulp_client_inst.unassociate_module_defaults, self.dst_repo, self.units)
        ]


class AssociateActionRpms(AssociateAction):
    TYPE = "packages"

    def get_actions(self, pulp_client_inst):
        actions = []
        for src_repo_id, units in self._map_src_repo_to_unit().items():
            actions.append(
                (
                    pulp_client_inst.associate_packages,
                    self._get_repo_obj(src_repo_id),
                    self.dst_repo,
                    units,
                )
            )

        return actions


class UnassociateActionRpms(PulpAction):
    TYPE = "packages"

    def get_actions(self, pulp_client_inst):
        return [(pulp_client_inst.unassociate_packages, self.dst_repo, self.units)]


def flatten_md_defaults_name_profiles(obj):
    """
    flatten the profiles of md_defaults unit and prepend name
    format: name:[key:profile,profile]:[key:profile]
    'ruby:[2.5:common,unique]'
    """
    result = obj.name
    for key in sorted(obj.profiles):
        result += ":[%s:%s]" % (key, ",".join(sorted(obj.profiles[key])))
    return result



from datetime import date
import logging
import re
import os

from collections import defaultdict, deque, namedtuple
from concurrent.futures import as_completed
from itertools import chain
from pubtools.pulplib import Client, Criteria, PublishOptions
from fastpurge import FastPurgeClient

import attr
import ubiconfig

from more_executors import Executors
from more_executors.futures import f_sequence, f_proxy, f_return
from ubipop._pulp_client import Pulp
from ubipop._utils import (
    AssociateActionModules,
    AssociateActionModuleDefaults,
    AssociateActionRpms,
    UnassociateActionModules,
    UnassociateActionModuleDefaults,
    UnassociateActionRpms,
    flatten_md_defaults_name_profiles,
)
from ._matcher import Matcher
from .ubi_manifest_client.client import Client as UbimClient
from ._cdn import Publisher

_LOG = logging.getLogger("ubipop")


class RepoMissing(Exception):
    pass


class ConfigMissing(Exception):
    pass


class PopulationSourceMissing(Exception):
    pass


RepoSet = namedtuple("RepoSet", ["rpm", "source", "debug"])


@attr.s
class RepoContent(object):
    binary_rpms = attr.ib()
    source_rpms = attr.ib()
    debug_rpms = attr.ib()
    modules = attr.ib()
    modulemd_defaults = attr.ib()


class UbiRepoSet(object):
    def __init__(self, input_repos, output_repos):
        self.in_repos = input_repos
        self.out_repos = output_repos

        self.packages = defaultdict(list)
        self.debug_rpms = defaultdict(list)
        self.modules = None
        self.module_defaults = defaultdict(list)
        self.source_rpms = defaultdict(list)

        self._ensure_repos_existence()

    def get_output_repos(self):
        repos = set([self.out_repos.rpm, self.out_repos.source])
        if self.out_repos.debug.result():
            repos.add(self.out_repos.debug)

        return repos

    def _ensure_repos_existence(self):
        fatal = False
        if not self.in_repos.rpm:
            fatal = True
            _LOG.error("Input Rpm repo does not exist")

        if not self.in_repos.source:
            fatal = True
            _LOG.error("Input Source repo does not exist")

        if not self.in_repos.debug:
            fatal = True
            _LOG.error("Input Debug repo does not exist")

        if not self.out_repos.rpm:
            fatal = True
            _LOG.error("Output Rpm repo does not exist")

        if not self.out_repos.source:
            fatal = True
            _LOG.error("Output Source repo does not exist")

        if not self.out_repos.debug:
            fatal = True
            _LOG.error("Output Debug repo does not exist")

        if fatal:
            raise RepoMissing()


class UbiPopulate(object):
    def __init__(
        self,
        pulp_hostname,
        pulp_auth,
        dry_run,
        ubiconfig_filename_list=None,
        ubiconfig_dir_or_url=None,
        verify=True,
        workers_count=4,
        output_repos=None,
        **kwargs
    ):
        # legacy client implemeted in this repo, it's expected to be replaced by pubtools.pulplib.Client
        self.pulp = self._make_pulp_client(pulp_hostname, pulp_auth, verify, Pulp)
        self._pulp_hostname = pulp_hostname
        self._pulp_auth = pulp_auth
        self._verify = verify
        self._pulp_client = None
        self.dry_run = dry_run
        self.output_repos = output_repos
        self._executor = Executors.thread_pool(max_workers=workers_count).with_retry()
        self._ubiconfig_list = None
        self._ubiconfig_filename_list = ubiconfig_filename_list
        self._ubiconfig_dir_or_url = ubiconfig_dir_or_url
        self._content_sets = kwargs.get("content_sets", None)
        self._repo_ids = kwargs.get("repo_ids", None)
        self._version = kwargs.get("version", None)
        self._content_set_regex = kwargs.get("content_set_regex", None)
        self._ubiconfig_map = None

        self._ubi_manifest_url = kwargs.get("ubi_manifest_url") or None

        self._publisher_args = {
            "edgerc": os.getenv("UBIPOP_EDGERC_CFG", "/etc/.edgerc"),
            "publish_options": {
                "clean": True,
            },
            "cdn_root": os.getenv("UBIPOP_CDN_ROOT", ""),
            "arl_templates": os.getenv("UBIPOP_ARL_TEMPLATES", "").split(","),
            "cert": (
                os.getenv("UBIPOP_CDN_CERT", ""),
                os.getenv("UBIPOP_CDN_KEY", ""),
            ),
            "verify": os.getenv("UBIPOP_CDN_CA_CERT", ""),
        }

    @property
    def pulp_client(self):
        if self._pulp_client is None:
            self._pulp_client = self._make_pulp_client(
                self._pulp_hostname, self._pulp_auth, self._verify, Client
            )
        return self._pulp_client

    def _make_pulp_client(self, url, auth, verify, client_klass):
        kwargs = {
            "verify": verify,
        }
        if os.path.isfile(auth[0]) and os.path.isfile(auth[1]):
            kwargs["cert"] = auth
        else:
            kwargs["auth"] = auth

        return client_klass("https://" + url, **kwargs)

    @property
    def ubiconfig_list(self):
        if self._ubiconfig_list is None:
            self._ubiconfig_list = self._load_ubiconfig()
        return self._ubiconfig_list

    def _load_ubiconfig(self):
        ubiconfig_dir_or_url = self._ubiconfig_dir_or_url
        filenames = self._ubiconfig_filename_list
        loader = ubiconfig.get_loader(ubiconfig_dir_or_url)
        ubi_conf_list = []

        if filenames:
            ubi_conf_list.extend(loader.load(filename) for filename in filenames)
        else:
            ubi_conf_list.extend(loader.load_all())

        return self._filter_ubi_conf_list(ubi_conf_list)

    @property
    def ubiconfig_map(self):
        if self._ubiconfig_map is None:
            self._ubiconfig_map = self._create_config_map()
        return self._ubiconfig_map

    def _filter_ubi_conf_list(self, config_list):
        """
        Reduces the list of UBI configurations to only those matching
        provided content sets and/or repo IDs. Config can be further filter
        by major version of or content_set_regex.
        """

        content_sets = self._content_sets
        repo_ids = self._repo_ids
        version = self._version
        content_set_regex = self._content_set_regex

        filtered_conf_list = []

        content_sets = content_sets or []
        repo_ids = repo_ids or []

        if not any([content_sets, repo_ids, version, content_set_regex]):
            # no filtering requested, return complete list of ubi config
            return config_list

        if content_set_regex:
            content_set_regex = re.compile(content_set_regex)

        if repo_ids:
            repos = [self.pulp_client.get_repository(repo_id) for repo_id in repo_ids]

            for repo in repos:
                content_sets.append(repo.content_set)

        for conf in config_list:
            # if requested, filter config by major version
            if version and not conf.version.startswith(version):
                continue

            for label in [
                conf.content_sets.rpm.input,
                conf.content_sets.rpm.output,
                conf.content_sets.srpm.input,
                conf.content_sets.srpm.output,
                conf.content_sets.debuginfo.input,
                conf.content_sets.debuginfo.output,
            ]:

                # matching by specific content sets takes precedence
                # or if empty content_set, content_set_regex - take the config immediately
                # we don't have any other filter to use
                if (
                    not any([content_sets, content_set_regex])
                    or label in content_sets
                    or (content_set_regex and re.search(content_set_regex, label))
                ):

                    filtered_conf_list.append(conf)
                    break

        return filtered_conf_list

    def _create_config_map(self):
        """Create a config map from self.ubiconfig_list, it has the form in:
        {
            "7.7":
                {
                    "config_filename1": config1,
                    "config_filename2": config2,
                    ...,
                },
            "8.1":
                {
                    "config_filename1": config1,
                    ...,
                },
            ....
        }
        """

        config_map = {}
        for config in self.ubiconfig_list:
            config_map.setdefault(config.version, {}).setdefault(
                config.file_name, config
            )

        return config_map

    def _get_config(self, repo, config):
        # get the right config file by ubi_config_version attr of a repository
        # if not found, try to fallback to the default version (major version)
        if not repo.ubi_config_version:
            raise ValueError("Repo: %s does not have ubi_config_version" % repo.id)

        _ubi_config_version = repo.ubi_config_version
        if _ubi_config_version not in self.ubiconfig_map:
            # if the config is missing, we need to use the default config branch
            _ubi_config_version = _ubi_config_version.split(".")[0]
        try:
            right_config = self.ubiconfig_map[_ubi_config_version][config.file_name]
        except KeyError:
            _LOG.error(
                "Config file %s missing from %s and default %s branches",
                config.file_name,
                repo.ubi_config_version,
                repo.ubi_config_version.split(".")[0],
            )
            raise ConfigMissing()

        return right_config

    def populate_ubi_repos(self):
        out_repos = set()
        used_content_sets = set()
        # since repos are searched by content sets, same repo could be searched and populated
        # multiple times, to avoid that, cache the content sets already used and skip the config
        # whose content sets are all in the cache
        repo_pairs_list = []
        ubi_binary_repos = []
        for config in sorted(self.ubiconfig_list, key=str):
            content_sets = [
                config.content_sets.rpm.output,
                config.content_sets.srpm.output,
                config.content_sets.debuginfo.output,
            ]

            to_use = [cs for cs in content_sets if cs not in used_content_sets]
            if to_use:
                for cs in to_use:
                    used_content_sets.add(cs)
            else:
                _LOG.debug(
                    "Skipping %s, since it's been used already", config.file_name
                )
                continue
            try:
                repo_pairs = self._get_ubi_repo_sets(config.content_sets.rpm.output)

            except RepoMissing:
                _LOG.warning("Skipping current content triplet, some repos are missing")
                continue

            repo_pairs_list.append((repo_pairs, config))
            ubi_binary_repos.extend(
                [repo_set.out_repos.rpm.id for repo_set in repo_pairs]
            )

        with UbimClient(self._ubi_manifest_url) as ubim_client:
            tasks = ubim_client.generate_manifest(ubi_binary_repos)
            tasks.result()

            with Publisher(**self._publisher_args) as publisher:
                self._run_ubi_population(
                    repo_pairs_list, out_repos, ubim_client, publisher
                )
                publisher.wait_publish_and_purge_cache()

        if self.output_repos:
            with open(self.output_repos, "w") as f:
                for repo in out_repos:
                    f.write(repo.id.strip() + "\n")

    def _run_ubi_population(
        self, repo_pairs_list, out_repos, ubim_client=None, publisher=None
    ):
        for repo_pairs, config in repo_pairs_list:
            for repo_set in repo_pairs:
                right_config = self._get_config(repo_set.out_repos.rpm, config)

                UbiPopulateRunner(
                    self.pulp,
                    self.pulp_client,
                    repo_set,
                    right_config,
                    self.dry_run,
                    self._executor,
                    ubim_client,
                    publisher,
                ).run_ubi_population()

                out_repos.update(repo_set.get_output_repos())

    def _get_ubi_repo_sets(self, ubi_binary_cs):
        """
        Searches for ubi repository triplet (binary rpm, srpm, debug) for
        one ubi config item and tries to determine their population sources
        (input repositories). Returns list UbiRepoSet objects that provides
        input and output repositories that are used for population process.
        """
        rpm_repos = self.pulp_client.search_repository(
            Criteria.and_(
                Criteria.with_field("notes.content_set", ubi_binary_cs),
                Criteria.with_field("ubi_population", True),
            )
        )

        ubi_repo_sets = []
        for out_rpm_repo in rpm_repos:
            out_source_repo = out_rpm_repo.get_source_repository()
            out_debug_repo = out_rpm_repo.get_debug_repository()

            in_rpm_repos = self._get_population_sources(out_rpm_repo)
            in_source_repos = self._get_population_sources(out_source_repo)
            in_debug_repos = self._get_population_sources(out_debug_repo)

            # we need to apply f_proxy(f_return()) for out_rpm_repo for keeping consistency
            # that all objects in out|in_repos are futures
            out_repos = (
                f_proxy(f_return(out_rpm_repo)),
                out_source_repo,
                out_debug_repo,
            )
            in_repos = (in_rpm_repos, in_source_repos, in_debug_repos)

            ubi_repo_sets.append(UbiRepoSet(RepoSet(*in_repos), RepoSet(*out_repos)))

        return ubi_repo_sets

    def _get_population_sources(self, out_repo):
        if (
            not hasattr(out_repo, "population_sources")
            or not out_repo.population_sources
        ):
            return []

        repos = [
            self.pulp_client.get_repository(repo_id)
            for repo_id in out_repo.population_sources
        ]

        return repos


class UbiPopulateRunner(object):
    def __init__(
        self,
        legacy_client,
        pulp_client,
        output_repo_set,
        ubiconfig_item,
        dry_run,
        executor,
        ubi_manifest_client=None,
        publisher=None,
    ):
        self.pulp = legacy_client
        self.pulp_client = pulp_client
        self.ubim_client = ubi_manifest_client

        self.repos = output_repo_set
        self.ubiconfig = ubiconfig_item
        self.dry_run = dry_run
        self._executor = executor
        self._publisher = publisher

    def _determine_pulp_actions(self, units, current, diff_f, extra_units=None):
        expected = list(units)
        if extra_units:
            expected += list(extra_units)
        to_associate = diff_f(expected, current)
        to_unassociate = diff_f(current, expected)
        return to_associate, to_unassociate

    def _get_pulp_actions_mds(self, modules, current):
        return self._determine_pulp_actions(
            modules, current, self._diff_modules_by_nsvca
        )

    def _get_pulp_actions_md_defaults(self, module_defaults, current):
        return self._determine_pulp_actions(
            module_defaults,
            current,
            self._diff_md_defaults_by_profiles,
        )

    def _get_pulp_actions_pkgs(self, pkgs, current, modular_pkgs):
        return self._determine_pulp_actions(
            pkgs, current, self._diff_packages_by_filename, modular_pkgs
        )

    def _get_pulp_actions_src_pkgs(self, pkgs, current, modular):
        """
        Get required pulp actions to make sure existing and desired source packages are in
        match.
        """
        uniq_srpms = {}

        all_pkgs = list(pkgs) + list(modular)

        # filter out packages that share same source rpm
        for pkg in all_pkgs:
            fn = pkg.filename or pkg.sourcerpm
            uniq_srpms[fn] = pkg

        src_pkgs = list(uniq_srpms.values())
        return self._determine_pulp_actions(
            src_pkgs, current, self._diff_packages_by_filename
        )

    def _get_pulp_actions(
        self,
        current_content,
        modular_binary,
        modular_debug,
        modular_source,
    ):
        """
        Determines expected pulp actions by comparing current content of output repos and
        expected content.

        Content that needs association: unit is in expected but not in current
        Content that needs unassociation: unit is in current but not in expected
        No action: unit is in current and in expected
        """

        modules_assoc, modules_unassoc = self._get_pulp_actions_mds(
            self.repos.modules, current_content.modules
        )
        md_defaults_assoc, md_defaults_unassoc = self._get_pulp_actions_md_defaults(
            self.repos.module_defaults, current_content.modulemd_defaults
        )

        rpms_assoc, rpms_unassoc = self._get_pulp_actions_pkgs(
            self.repos.packages, current_content.binary_rpms, modular_binary
        )
        srpms_assoc, srpms_unassoc = self._get_pulp_actions_src_pkgs(
            self.repos.source_rpms, current_content.source_rpms, modular_source
        )

        debug_assoc = None
        debug_unassoc = None
        if current_content.debug_rpms:
            debug_assoc, debug_unassoc = self._get_pulp_actions_pkgs(
                self.repos.debug_rpms, current_content.debug_rpms, modular_debug
            )

        associations = (
            AssociateActionModules(
                modules_assoc, self.repos.out_repos.rpm, self.repos.in_repos.rpm
            ),
            AssociateActionRpms(
                rpms_assoc, self.repos.out_repos.rpm, self.repos.in_repos.rpm
            ),
            AssociateActionRpms(
                srpms_assoc, self.repos.out_repos.source, self.repos.in_repos.source
            ),
            AssociateActionRpms(
                debug_assoc, self.repos.out_repos.debug, self.repos.in_repos.debug
            ),
        )

        unassociations = (
            UnassociateActionModules(modules_unassoc, self.repos.out_repos.rpm),
            UnassociateActionRpms(rpms_unassoc, self.repos.out_repos.rpm),
            UnassociateActionRpms(srpms_unassoc, self.repos.out_repos.source),
            UnassociateActionRpms(debug_unassoc, self.repos.out_repos.debug),
        )

        mdd_association = AssociateActionModuleDefaults(
            md_defaults_assoc, self.repos.out_repos.rpm, self.repos.in_repos.rpm
        )

        mdd_unassociation = UnassociateActionModuleDefaults(
            md_defaults_unassoc, self.repos.out_repos.rpm
        )

        return associations, unassociations, mdd_association, mdd_unassociation

    def _diff_modules_by_nsvca(self, modules_1, modules_2):
        return self._diff_lists_by_attr(modules_1, modules_2, "nsvca")

    def _diff_md_defaults_by_profiles(self, module_defaults_1, module_defaults_2):
        return self._diff_lists_by_attr(
            module_defaults_1, module_defaults_2, flatten_md_defaults_name_profiles
        )

    def _diff_packages_by_filename(self, packages_1, packages_2):
        return self._diff_lists_by_attr(packages_1, packages_2, "filename")

    def _diff_lists_by_attr(self, list_1, list_2, attr_or_func):
        def diff_attr(obj):
            if callable(attr_or_func):
                return attr_or_func(obj)
            return getattr(obj, attr_or_func)

        attrs_list_2 = [diff_attr(obj) for obj in list_2]
        diff = [obj for obj in list_1 if diff_attr(obj) not in attrs_list_2]

        return diff

    def _search_expected_modulemd_defaults(self, modulemd_defaults):
        criteria_values = [(unit.name,) for unit in modulemd_defaults]
        fields = ("name",)
        or_criteria = Matcher.create_or_criteria(fields, criteria_values)
        return f_proxy(
            self._executor.submit(
                Matcher.search_modulemd_defaults, or_criteria, self.repos.in_repos.rpm
            )
        )

    def run_ubi_population(self):
        current_content = self._get_current_content()

        modular_rpms = []
        modular_debug_rpms = []
        modular_source_rpms = []

        # start async querying for modulemds and modular and non-modular packages
        binary_manifest = self.ubim_client.get_manifest(self.repos.out_repos.rpm.id)
        debug_manifest = self.ubim_client.get_manifest(self.repos.out_repos.debug.id)
        source_manifest = self.ubim_client.get_manifest(self.repos.out_repos.source.id)
        self.repos.modules = binary_manifest.modules
        self.repos.module_defaults = self._search_expected_modulemd_defaults(
            binary_manifest.modulemd_defaults
        )
        self.repos.packages = binary_manifest.packages
        self.repos.debug_rpms = debug_manifest.packages
        self.repos.source_rpms = source_manifest.packages

        (
            associations,
            unassociations,
            mdd_association,
            mdd_unassociation,
        ) = self._get_pulp_actions(
            current_content,
            modular_rpms,
            modular_debug_rpms,
            modular_source_rpms,
        )

        if self.dry_run:
            self.log_curent_content(current_content)
            self.log_pulp_actions(
                associations + (mdd_association,),
                unassociations + (mdd_unassociation,),
            )
        else:
            fts = []
            fts.extend(self._associate_unassociate_units(associations + unassociations))
            # wait for associate/unassociate tasks
            self._wait_pulp(fts)

            self._associate_unassociate_md_defaults(
                (mdd_association,), (mdd_unassociation,)
            )

            self._publish_out_repos()

    def _associate_unassociate_units(self, action_list):
        fts = []
        for action in action_list:
            if action.units:
                fts.extend(
                    [self._executor.submit(*a) for a in action.get_actions(self.pulp)]
                )

        return fts

    def _associate_unassociate_md_defaults(self, action_md_ass, action_md_unass):
        """
        Unassociate old module defaults units first, wait until done and
        then start new units association
        """
        fts_unass = self._associate_unassociate_units(action_md_unass)
        self._wait_pulp(fts_unass)

        fts_ass = self._associate_unassociate_units(action_md_ass)
        self._wait_pulp(fts_ass)

    def _wait_pulp(self, futures):
        # wait for pulp tasks from futures
        for ft in as_completed(futures):
            tasks = ft.result()
            if tasks:
                self.pulp.wait_for_tasks(tasks)

    def log_curent_content(self, current_content):
        _LOG.info("Current modules in repo: %s", self.repos.out_repos.rpm.id)
        for module in current_content.modules:
            _LOG.info(module.nsvca)

        _LOG.info("Current module_defaults in repo: %s", self.repos.out_repos.rpm.id)
        for md_d in current_content.modulemd_defaults:
            _LOG.info("module_defaults: %s, profiles: %s", md_d.name, md_d.profiles)

        _LOG.info("Current rpms in repo: %s", self.repos.out_repos.rpm.id)
        for rpm in current_content.binary_rpms:
            _LOG.info(rpm.filename)

        _LOG.info("Current srpms in repo: %s", self.repos.out_repos.source.id)
        for rpm in current_content.source_rpms:
            _LOG.info(rpm.filename)

        if self.repos.out_repos.debug:
            _LOG.info("Current rpms in repo: %s", self.repos.out_repos.debug.id)
            for rpm in current_content.debug_rpms:
                _LOG.info(rpm.filename)

    def log_pulp_actions(self, associations, unassociations):
        for item in associations:
            if item.units:
                for unit in item.units:
                    _LOG.info(
                        "Would associate %s from %s to %s",
                        unit,
                        unit.associate_source_repo_id,
                        item.dst_repo.id,
                    )

            else:
                _LOG.info(
                    "No association expected for %s from %s to %s",
                    item.TYPE,
                    [r.id for r in item.src_repos],
                    item.dst_repo.id,
                )

        for item in unassociations:
            if item.units:
                for unit in item.units:
                    _LOG.info("Would unassociate %s from %s", unit, item.dst_repo.id)
            else:
                _LOG.info(
                    "No unassociation expected for %s from %s",
                    item.TYPE,
                    item.dst_repo.id,
                )

    def _get_current_content(self):
        """
        Gather current content of output repos
        """
        criteria = [Criteria.true()]
        current_modulemds = f_proxy(
            self._executor.submit(
                Matcher.search_modulemds, criteria, [self.repos.out_repos.rpm]
            )
        )
        current_modulemd_defaults = f_proxy(
            self._executor.submit(
                Matcher.search_modulemd_defaults, criteria, [self.repos.out_repos.rpm]
            )
        )
        current_rpms = f_proxy(
            self._executor.submit(
                Matcher.search_rpms, criteria, [self.repos.out_repos.rpm]
            )
        )
        current_srpms = f_proxy(
            self._executor.submit(
                Matcher.search_srpms, criteria, [self.repos.out_repos.source]
            )
        )

        if self.repos.out_repos.debug.result():
            current_debug_rpms = f_proxy(
                self._executor.submit(
                    Matcher.search_rpms, criteria, [self.repos.out_repos.debug]
                )
            )
        else:
            current_debug_rpms = f_proxy(f_return([]))

        current_content = RepoContent(
            current_rpms,
            current_srpms,
            current_debug_rpms,
            current_modulemds,
            current_modulemd_defaults,
        )
        return current_content

    def _publish_out_repos(self):
        to_publish = []

        for repo in (
            self.repos.out_repos.rpm,
            self.repos.out_repos.debug,
            self.repos.out_repos.source,
        ):
            if repo.result():
                to_publish.append(repo)
        self._publisher.enqueue(*to_publish)



import os
import logging

from pubtools.pulplib import Criteria
from more_executors.futures import f_flat_map, f_return, f_sequence, f_proxy
from more_executors import Executors

BATCH_SIZE = int(os.getenv("UBIPOP_BATCH_SIZE", "250"))
# need to set significantly lower batches for general rpm search
# otherwise db may very likely hit OOM error.
BATCH_SIZE_RPM = int(os.getenv("UBIPOP_BATCH_SIZE_RPM", "15"))

_LOG = logging.getLogger("ubipop.matcher")


class UbiUnit(object):
    """
    Wrapping class of model classes (*Unit) of pubtools.pulplib.
    """

    def __init__(self, unit, src_repo_id):
        self._unit = unit
        self.associate_source_repo_id = src_repo_id

    def __getattr__(self, name):
        return getattr(self._unit, name)

    def __str__(self):
        return str(self._unit)

    # TODO make this return hash of self._unit if possible in future
    # it should help us with not adding the same units into sets
    # that differ with associate_source_repo_id attr only
    # currently some *Unit classes from pulplib are not hashable
    # def __hash__(self):
    #    return hash(self._unit)


class Matcher(object):
    """
    Generic class can be used for matching units in pulp required for ubipop.
    This class should be subclasses and run() method needs to be implemented.
    """

    def __init__(self, input_repos, ubi_config, workers=8):
        self._input_repos = input_repos
        self._ubi_config = ubi_config
        # executor for this class, not adding retries because for pulp
        # we use executor from pulplib
        self._executor = Executors.thread_pool(max_workers=workers)

        self.binary_rpms = None
        self.debug_rpms = None
        self.source_rpms = None

    def run(self):
        """
        This method needs to be implemented in subclasses and should
        include only async calls that will set public attributes of the class
        and immediately return self.
        """
        raise NotImplementedError

    @classmethod
    def _search_units(
        cls, repo, criteria_list, content_type_id, batch_size_override=None
    ):
        """
        Search for units of one content type associated with given repository by criteria.
        """
        units = set()
        batch_size = batch_size_override or BATCH_SIZE

        def handle_results(page):
            for unit in page.data:
                unit = UbiUnit(unit, repo.id)
                units.add(unit)
            if page.next:
                return f_flat_map(page.next, handle_results)
            return f_return(units)

        criteria_split = []

        for start in range(0, len(criteria_list), batch_size):
            criteria_split.append(criteria_list[start : start + batch_size])
        fts = []

        for criteria_batch in criteria_split:
            _criteria = Criteria.and_(
                Criteria.with_field("content_type_id", content_type_id),
                Criteria.or_(*criteria_batch),
            )

            page_f = repo.search_content(_criteria)
            handled_f = f_flat_map(page_f, handle_results)

            fts.append(handled_f)

        return f_flat_map(f_sequence(fts), flatten_list_of_sets)

    @classmethod
    def create_or_criteria(cls, fields, values):
        # fields - list/tuple of fields [field1, field2]
        # values - list of tuples [(field1 value, field2 value), ...]
        # creates criteria for pulp query in a following way
        # one tuple in values uses AND logic
        # each criteria for one tuple are agregated by to or_criteria list
        or_criteria = []

        for val_tuple in values:
            inner_and_criteria = []
            if len(val_tuple) != len(fields):
                raise ValueError
            for index, field in enumerate(fields):

                inner_and_criteria.append(Criteria.with_field(field, val_tuple[index]))

            or_criteria.append(Criteria.and_(*inner_and_criteria))

        return or_criteria

    @classmethod
    def _search_units_per_repos(
        cls, or_criteria, repos, content_type, batch_size_override=None
    ):
        units = []
        for repo in repos:
            units.append(
                cls._search_units(
                    repo,
                    or_criteria,
                    content_type,
                    batch_size_override=batch_size_override,
                )
            )

        return f_proxy(f_flat_map(f_sequence(units), flatten_list_of_sets))

    @classmethod
    def search_rpms(cls, or_criteria, repos, batch_size_override=None):
        return cls._search_units_per_repos(
            or_criteria,
            repos,
            content_type="rpm",
            batch_size_override=batch_size_override,
        )

    @classmethod
    def search_srpms(cls, or_criteria, repos, batch_size_override=None):
        return cls._search_units_per_repos(
            or_criteria,
            repos,
            content_type="srpm",
            batch_size_override=batch_size_override,
        )

    @classmethod
    def search_modulemds(cls, or_criteria, repos):
        return cls._search_units_per_repos(or_criteria, repos, content_type="modulemd")

    @classmethod
    def search_modulemd_defaults(cls, or_criteria, repos):
        return cls._search_units_per_repos(
            or_criteria, repos, content_type="modulemd_defaults"
        )


def flatten_list_of_sets(list_of_sets):
    out = set()
    for one_set in list_of_sets:
        out |= one_set

    return f_return(out)



# inspired by CDN client and cache purge implementation
# in https://github.com/release-engineering/pubtools-pulp/
import logging
import os
import re
import threading
from collections import namedtuple
from concurrent.futures import as_completed

import requests
from fastpurge import FastPurgeClient
from more_executors import Executors
from more_executors.futures import f_map, f_return, f_sequence
from pubtools.pulplib import PublishOptions

LOG = logging.getLogger("ubipop")


HeaderPair = namedtuple("HeaderPair", ["request", "response"])
TtlConfig = namedtuple("TtlConfig", ["regex", "ttl"])


class Publisher:
    def __init__(self, publish_options=None, edgerc=None, cdn_root=None, **kwargs):
        """Create a new Publisher object.
        Which can be used to enqueuing repository published with CDN cache purge.

        Arguments:
            publish_options (dict)
                supported args for PublishOptions class

            edgerc (str|dict)
                path to edgerc config or dict with config

            cdn_root (str)
                root url of CDN service (e.g. https://cdn.example.com)

            kwargs:
                other args used for CdnClient (e.g. arl_templates, cert, verify)

        """
        self._cdn_client = None
        self._cdn_cache_purger = None

        self._publish_queue = set()

        publish_options = publish_options if publish_options else {}
        self._publish_options = PublishOptions(**publish_options)

        self._edgerc = edgerc
        self._cdn_client_args = {
            "url": cdn_root,
        }
        for arg in ("cert", "verify", "arl_templates"):
            self._cdn_client_args[arg] = kwargs.get(arg)

        self._cdn_root = cdn_root

    def __enter__(self):
        return self

    def __exit__(self, *exc_details):
        if self._cdn_client:
            self._cdn_client.__exit__(*exc_details)
        if self._cdn_cache_purger:
            self._cdn_cache_purger.__exit__(*exc_details)

    @property
    def cdn_client(self):
        if not all(
            [self._cdn_client_args["url"], self._cdn_client_args["arl_templates"]]
        ):
            return None

        if not self._cdn_client:
            self._cdn_client = CdnClient(**self._cdn_client_args)
        return self._cdn_client

    @property
    def cdn_cache_purger(self):
        if not self._cdn_root:
            return None

        if not self._cdn_cache_purger:
            self._cdn_cache_purger = CdnCachePurger(self._edgerc)

        return self._cdn_cache_purger

    def enqueue(self, *repos):
        def _enqueue(repo):
            ft = repo.publish(self._publish_options)
            ft = f_map(ft, lambda _: repo)
            self._publish_queue.add(ft)

        _ = [_enqueue(r) for r in repos if r.result() is not None]

    def wait_publish_and_purge_cache(self):
        f_sequence(self._publish_queue).result()
        LOG.info("Publish finished")
        if self.cdn_cache_purger:
            LOG.info("CDN cache purge started")
            self._purge_cache().result()
            LOG.info("CDN cache purge finished")
        else:
            LOG.info("CDN cache purge disabled.")

    def _purge_cache(self):
        purges = []

        for repo in as_completed(self._publish_queue):
            _repo = repo.result()
            if _repo.relative_url:
                for mutable_url in _repo.mutable_urls:
                    relative_mutable_url = os.path.join(_repo.relative_url, mutable_url)

                    url = os.path.join(self._cdn_root, relative_mutable_url)
                    purges.append(f_return(url))
                    if self.cdn_client:
                        arls_ft = self.cdn_client.get_arl_for_path(relative_mutable_url)
                        purges.extend(arls_ft)

        return f_map(f_sequence(purges), self.cdn_cache_purger.purge_by_url)


class CdnCachePurger:
    def __init__(self, edgerc):
        """Create a new CDN cache purger client.

        Arguments:
            edgerc (str|dict)
                Path to edgerc config or dict containing the config.
        """
        self._edgerc = edgerc
        self._fastpurge_client = None

    def __enter__(self):
        return self

    def __exit__(self, *exc_details):
        if self._fastpurge_client:
            self._fastpurge_client.__exit__(*exc_details)

    @property
    def fastpurge_client(self):
        if not self._fastpurge_client:
            self._fastpurge_client = FastPurgeClient(auth=self._edgerc)

        return self._fastpurge_client

    def purge_by_url(self, urls):
        return self.fastpurge_client.purge_by_url(urls)


class CdnClient:
    # Client for requesting special headers from CDN service.

    # Default number of request thread modifiable by an env variable.
    # This is not a documented/supported feature of the library.
    _REQUEST_THREADS = int(os.environ.get("CDN_REQUEST_THREADS", "4"))
    _ATTEMPTS = int(os.environ.get("CDN_RETRY_ATTEMPTS", "9"))
    _SLEEP = float(os.environ.get("CDN_RETRY_SLEEP", "1.0"))
    _EXPONENT = float(os.environ.get("CDN_RETRY_EXPONENT", "3.0"))
    _MAX_SLEEP = float(os.environ.get("CDN_RETRY_MAX_SLEEP", "120.0"))

    TTL_REGEX = re.compile(r".*/(\d+[smhd])/.*")
    CACHE_KEY_HEADER = HeaderPair("akamai-x-get-cache-key", "X-Cache-Key")

    def __init__(self, url, arl_templates=None, max_retry_sleep=_MAX_SLEEP, **kwargs):
        """Create a new CDN client.

        Arguments:
            url (str)
                Base URL of CDN
            arl_templates List[str]
                Templates used for ARL generation
            max_retry_sleep (float)
                Max number of seconds to sleep between retries.
                Mainly provided so that tests can reduce the time needed to retry.
            kwargs
                Remaining arguments are used to initialize the requests.Session()
                used within this class (e.g. "verify", "cert").
        """
        self._url = url
        self._arl_templates = arl_templates
        self._tls = threading.local()

        retry_args = {
            "max_sleep": max_retry_sleep,
            "max_attempts": CdnClient._ATTEMPTS,
            "sleep": CdnClient._SLEEP,
            "exponent": CdnClient._EXPONENT,
        }
        self._session_attrs = kwargs
        self._executor = (
            Executors.thread_pool(name="cdn-client", max_workers=self._REQUEST_THREADS)
            .with_map(self._check_http_response)
            .with_retry(**retry_args)
        )

    def __enter__(self):
        return self

    def __exit__(self, *exc_details):
        self._executor.__exit__(*exc_details)

    @staticmethod
    def _check_http_response(response):
        response.raise_for_status()
        return response

    @property
    def _session(self):
        if not hasattr(self._tls, "session"):
            self._tls.session = requests.Session()
            for key, value in self._session_attrs.items():
                setattr(self._tls.session, key, value)
        return self._tls.session

    def _head(self, *args, **kwargs):
        return self._session.head(*args, **kwargs)

    def _on_failure(self, header, exception):
        LOG.error("Requesting header %s failed: %s", header, exception)
        raise exception

    def _get_headers_for_path(self, path, headers):
        url = os.path.join(self._url, path)

        LOG.info("Getting headers %s for %s", list(headers.values()), url)

        out = self._executor.submit(self._head, url, headers=headers)
        out = f_map(
            out,
            fn=lambda resp: resp.headers,
            error_fn=lambda ex: self._on_failure(list(headers.values()), ex),
        )

        return out

    def _get_ttl(self, path):
        headers = {"Pragma": self.CACHE_KEY_HEADER.request}

        out = self._get_headers_for_path(path, headers)

        def _parse_ttl(value):
            parsed = re.match(
                self.TTL_REGEX, value.get(self.CACHE_KEY_HEADER.response) or ""
            )
            return parsed.group(1) if parsed else None

        return f_map(out, _parse_ttl)

    def _is_valid_template(self, template):
        return all(["{ttl}" in template, "{path}" in template])

    def get_arl_for_path(self, path):
        """Get ARL for particular path using provided templates.
        This method generates ARLs for given path according to
        provided ARL templates. TTL value is requested from CDN
        special headers.

        If value of TTL cannot be fetched from CDN service,
        we fallback to hardcoded values.

        Arguments:
            path (str)
                Relative path/URL (e.g. content/foo/bar/repomd.xml).
        Returns:
            List[Future]
                A list of futures holding formatted ARLs.
        """

        def _format_template(ttl, template, path):
            ttl = f_map(ttl, fn=lambda x: x, error_fn=lambda _: ttl_for_path(path))
            return f_map(ttl, lambda x: template.format(ttl=x, path=path))

        out = []
        ttl_ft = self._get_ttl(path)

        for item in self._arl_templates or []:
            if self._is_valid_template(item):
                out.append(_format_template(ttl_ft, item, path))
        return out


# ordering of items matters as it's used as priority
CDN_TTL_CONFIG = (
    TtlConfig(re.compile(r"/repodata/.*\.xml$"), "4h"),
    TtlConfig(re.compile(r".*/ostree/repo/refs/heads/.*/(base|standard)$"), "10m"),
    TtlConfig(re.compile(r"(/PULP_MANIFEST$|/listing$|/repodata/)"), "10m"),
    TtlConfig(re.compile(r"/$"), "4h"),
)

DEFAULT_TTL = "30d"


def ttl_for_path(path):
    out = DEFAULT_TTL
    for regex, ttl in CDN_TTL_CONFIG:
        if regex.search(path):
            out = ttl
            break

    return out



from setuptools import setup, find_packages


def get_description():
    return "Tool for populating ubi repositories"


def get_long_description():
    with open("README.md") as f:
        text = f.read()

    # Long description is everything after README's initial heading
    idx = text.find("\n\n")
    return text[idx:]


def get_requirements():
    with open("requirements.txt") as f:
        return f.read().splitlines()


setup(
    name="ubi-population-tool",
    version="0.14.0",
    license="GNU General Public License",
    author="",
    author_email="",
    description=get_description(),
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    url="https://github.com/release-engineering/ubi-population-tool",
    install_requires=get_requirements(),
    python_requires=">=3.7",
    packages=find_packages(exclude=["tests", "tests.*"]),
    entry_points={
        "console_scripts": [
            "ubipop = ubipop.cli:entry_point",
        ]
    },
)



import argparse
import logging
import sys

import ubipop

DEFAULT_LOG_FMT = "%(asctime)s [%(levelname)-8s] %(message)s"
DEFAULT_DATE_FMT = "%Y-%m-%d %H:%M:%S %z"

_LOG = logging.getLogger("ubipop")
_LOG.setLevel(logging.DEBUG)


def parse_args(args):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "input", action="store", nargs="*", help="path to ubi config file"
    )
    parser.add_argument(
        "--content-sets",
        action="store",
        nargs="+",
        type=str,
        required=False,
        help="content set labels from which to source ubi config",
    )
    parser.add_argument(
        "--repo-ids",
        action="store",
        nargs="+",
        type=str,
        required=False,
        help="repo IDs from which to source ubi config",
    )
    parser.add_argument(
        "--conf-src",
        action="store",
        required=False,
        help="source of ubi config, directory or url",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        default=False,
        help="if True, print pulp actions",
    )
    parser.add_argument(
        "--pulp-hostname", action="store", required=True, help="hostname of pulp_server"
    )
    parser.add_argument(
        "--insecure",
        action="store_true",
        default=False,
        help="use insecure connection to pulp server",
    )
    parser.add_argument("--user", action="store", required=False, help="pulp user")
    parser.add_argument(
        "--password", action="store", required=False, help="pulp password"
    )
    parser.add_argument(
        "--cert",
        action="store",
        required=False,
        help="path to to user cert for pulp authentication",
    )
    parser.add_argument(
        "--key",
        action="store",
        required=False,
        help="path to to user key for pulp authentication",
    )
    parser.add_argument(
        "--workers",
        action="store",
        type=int,
        default=4,
        help="Number of workers for parallel execution",
    )
    parser.add_argument(
        "--output-repos",
        action="store",
        required=False,
        help="Path to output file."
        "If provided, file containing repo ids of all out repos is created.",
    )
    parser.add_argument(
        "--version",
        action="store",
        required=False,
        help="Major version of ubi content set to be populated.",
    )
    parser.add_argument(
        "--content-set-regex",
        action="store",
        required=False,
        help="Regular expression of ubi content set to be populated.",
    )
    parser.add_argument(
        "--ubi-manifest-url",
        action="store",
        required=True,
        help="URL of ubi-manifest service",
    )

    parsed = parser.parse_args(args)

    auth_err_msg = "Provide --user and --password options or --cert and --key"
    if all((parsed.user, parsed.password, parsed.cert, parsed.key)):
        parser.error(auth_err_msg)

    auth = (None, None)
    if parsed.user and parsed.password:
        auth = (parsed.user, parsed.password)
    elif parsed.cert and parsed.key:
        auth = (parsed.cert, parsed.key)

    if not all(auth):
        parser.error(auth_err_msg)

    return parser.parse_args(args), auth


FOREIGN_LOGGERS = ("pubtools.pulplib", "fastpurge", "ubiconfig")


def main(args):
    logging.basicConfig(format=DEFAULT_LOG_FMT, datefmt=DEFAULT_DATE_FMT)
    for logger in FOREIGN_LOGGERS:
        logging.getLogger(logger).setLevel(logging.INFO)

    opts, auth = parse_args(args)

    ubipop.UbiPopulate(
        opts.pulp_hostname,
        auth,
        opts.dry_run,
        opts.input,
        opts.conf_src,
        not opts.insecure,
        opts.workers,
        opts.output_repos,
        content_sets=opts.content_sets,
        repo_ids=opts.repo_ids,
        version=opts.version,
        content_set_regex=opts.content_set_regex,
        ubi_manifest_url=opts.ubi_manifest_url,
    ).populate_ubi_repos()


def entry_point():
    main(sys.argv[1:])


if __name__ == "__main__":
    entry_point()
